import miney

# todo: a single look attribute (with combined horizontal/vertical) with degrees unit.
# todo: fly
# todo: set modes creative/survival -> Not possible without installed minetest mods


class Player:
    """
    A player of the minetest server.
    """
    def __init__(self, minetest: miney.Minetest, name):
        """
        Initialize the player object.

        :param minetest: Parent minetest object
        :param name: Player name
        """
        self.mt = minetest
        self.name = name
        
        # get user data: password hash, last login, privileges
        data = self.mt.lua.run("return minetest.get_auth_handler().get_auth('{}')".format(self.name))
        if data and all(k in data for k in ("password", "last_login", "privileges")):  # if we have all keys
            self.password = data["password"]
            self.last_login = data["last_login"]
            self.privileges = data["privileges"]
        else:
            raise miney.PlayerInvalid("There is no player with that name")

    def __repr__(self):
        return '<minetest player "{}">'.format(self.name)

    @property
    def is_online(self) -> bool:
        """
        Returns the online status of this player.

        :return: True/False
        """
        # TODO: Better check without provoke a lua error
        try:
            if self.name == self.mt.lua.run(
                    "return minetest.get_player_by_name('{}'):get_player_name()".format(self.name)):
                return True
        except miney.LuaError:
            return False

    @property
    def position(self) -> dict:
        """
        Get the players current position.

        :return: A dict with x,y,z keys: {x: 0, y:1, z:2}
        """
        try:
            return self.mt.lua.run("return minetest.get_player_by_name('{}'):get_pos()".format(self.name))
        except miney.LuaError:
            raise miney.PlayerOffline("The player has no position, he could be offline")

    @position.setter
    def position(self, values: dict):
        """
        Set player position
        :param values:
        :return:
        """
        if all(k in values for k in ("x", "y", "z")):  # if we have all keys
            self.mt.lua.run(
                "return minetest.get_player_by_name('{}'):set_pos({{x = {}, y = {}, z = {}}})".format(
                    self.name,
                    values["x"],
                    values["y"],
                    values["z"],
                )
            )
        else:
            raise miney.NoValidPosition(
                "A valid position need x,y,z values in an dict ({\"x\": 12, \"y\": 70, \"z\": 12}).")

    @property
    def speed(self) -> int:
        """
        Get or set the players speed. Default is 1.

        :return: Float
        """
        return self.mt.lua.run(
            "return minetest.get_player_by_name('{}'):get_physics_override()".format(self.name))["speed"]

    @speed.setter
    def speed(self, value: int):
        self.mt.lua.run(
            "return minetest.get_player_by_name('{}'):set_physics_override({{speed = {}}})".format(self.name, value))

    @property
    def jump(self):
        """
        Get or set the players jump height. Default is 1.

        :return: Float
        """
        return self.mt.lua.run(
            "return minetest.get_player_by_name('{}'):get_physics_override()".format(self.name))["jump"]

    @jump.setter
    def jump(self, value):
        self.mt.lua.run(
            "return minetest.get_player_by_name('{}'):set_physics_override({{jump = {}}})".format(self.name, value))

    @property
    def gravity(self):
        """
        Get or set the players gravity. Default is 1.

        :return: Float
        """
        return self.mt.lua.run(
            "return minetest.get_player_by_name('{}'):get_physics_override()".format(self.name))["gravity"]

    @gravity.setter
    def gravity(self, value):
        self.mt.lua.run(
            "return minetest.get_player_by_name('{}'):set_physics_override({{gravity = {}}})".format(self.name, value))

    @property
    def look_vertical_rad(self):
        """
        Get and set pitch in radians. Angle ranges between -pi/2 (~-1.563) and pi/2 (~1.563), which are straight
        up and down respectively.

        :return: Pitch in radians
        """
        return self.mt.lua.run("return minetest.get_player_by_name('{}'):get_look_vertical()".format(self.name))

    @look_vertical_rad.setter
    def look_vertical_rad(self, value):
        self.mt.lua.run("return minetest.get_player_by_name('{}'):set_look_vertical({})".format(self.name, value))

    @property
    def look_horizontal_rad(self):
        """
        Get and set yaw in radians. Angle is counter-clockwise from the +z direction.

        :return: Pitch in radians
        """
        return self.mt.lua.run("return minetest.get_player_by_name('{}'):get_look_horizontal()".format(self.name))

    @look_horizontal_rad.setter
    def look_horizontal_rad(self, value):
        self.mt.lua.run("return minetest.get_player_by_name('{}'):set_look_horizontal({})".format(self.name, value))



import socket
import json
from typing import Dict
import miney


# TODO: Implement Secure Remote Password protocol to replace plain passwords
# TODO: set_node with simple nodename or detailed dict with parameters
# TODO: bulk node setter
# TODO: vector node setter
# TODO: backup/restore ability
# TODO: Sending long strings to mineysocket


class Minetest:
    """__init__([server, playername, password, [port]])
    The Minetest server object. All other object are created from here. By creating an object you connect to Minetest.

    **Parameters aren't required, if you run minetest on the same computer.**

    *If you connect over LAN or Internet to a Minetest server with installed mineysocket, you have to provide a valid
    playername with a password:*

    ::

        mt = Minetest("192.168.0.2", "MyNick", "secret_password")

    Account creation is done by starting Minetest and connect to a server with a username
    and password. https://wiki.minetest.net/Getting_Started#Play_Online

    :param str server: IP or DNS name of an minetest server with installed apisocket mod
    :param str playername: A name to identify yourself to the server.
    :param str password: Your password
    :param int port: The apisocket port, defaults to 29999
    """
    def __init__(self, server: str = "127.0.0.1", playername: str = "Player", password: str = "", port: int = 29999):
        """
        Connect to the minetest server.

        :param server: IP or DNS name of an minetest server with installed apisocket mod
        :param port: The apisocket port, defaults to 29999
        """
        self.server = server
        self.port = port
        self.playername = playername
        self.password = password
        self.connection = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.connection.settimeout(2.0)
        self.event_queue = []  # List for collected but unprocessed events
        self.result_queue = {}  # List for unprocessed results

        # objects representing local properties
        self._chat: miney.chat.Chat = miney.Chat(self)
        self._node: miney.node.Node = miney.Node(self)
        self._lua: miney.lua.Lua = miney.Lua(self)

        self._authenticate()

    def _authenticate(self):
        """
        Authenticate to mineysocket.

        :return: None
        """
        # authenticate
        self.send({"playername": self.playername, "password": self.password})
        result = self.receive("auth")

        if "auth_ok" not in result:
            raise miney.AuthenticationError("Wrong playername or password")
        else:
            self.clientid = result[1]  # Clientid = <IP>:<Port>

    def send(self, data: Dict):
        """
        Send json objects to the miney-socket.

        :param data:
        :return:
        """
        self.connection.sendto(str.encode(json.dumps(data) + "\n"), (self.server, self.port))

    def receive(self, result_id: str = None, timeout: float = None):
        """
        Receive data and events from minetest.

        With an optional result_id this function waits for a result with that id by call itself until the right result
        was received. **If lua.run() was used this is not necessary, cause miney already takes care of it.**

        With the optional timeout the blocking waiting time for data can be changed.

        Example to receive and print all events::

            from miney import Minetest
            mt = Minetest()

            while True:
                print("Event received:", mt.receive())

        :param str result_id: Wait for this result id
        :param float timeout: Block further execution until data received or timeout in seconds is over.
        :return:
        """

        def format_result(result_data):
            if type(result_data["result"]) in (list, dict):
                if len(result_data["result"]) == 0:
                    return
                if len(result_data["result"]) == 1:  # list with single element doesn't needs a list
                    return result_data["result"][0]
                if len(result_data["result"]) > 1:
                    return tuple(result_data["result"])
            else:
                return result_data["result"]

        # Check if we have to return something received earlier
        if result_id in self.result_queue:
            result = format_result(self.result_queue[result_id])
            del self.result_queue[result_id]
            return result
        elif not result_id and len(self.event_queue):
            result = self.event_queue[0]
            del self.event_queue[0]
            return result

        # Set a new timeout to prevent long waiting for a timeout
        if timeout:
            self.connection.settimeout(timeout)

        try:
            # receive the raw data and try to decode json
            data_buffer = b""
            while "\n" not in data_buffer.decode():
                data_buffer = data_buffer + self.connection.recv(4096)
            data = json.loads(data_buffer.decode())
        except socket.timeout:
            return False

        # process data
        if "result" in data:
            if result_id:  # do we need a specific result?
                if data["id"] == result_id:  # we've got the result we wanted
                    return format_result(data)
            # We store this for later processing
            self.result_queue[data["id"]] = data
        elif "error" in data:
            if data["error"] == "authentication error":
                # maybe a server restart or timeout. We just reauthenticate.
                self._authenticate()
                raise miney.SessionReconnected()
            else:
                raise miney.LuaError("Lua-Error: " + data["error"])
        elif "event" in data:
            # return event only if we don't want a specific result
            if not result_id:
                return data
            else:
                # We collect and store a event for later processing
                self.event_queue.append(data)

        # if we don't got our result we have to receive again
        if result_id:
            self.receive(result_id)

    @property
    def chat(self):
        """
        Object with chat functions.

        Example::

            from miney import Minetest
            mt = Minetest()
            mt.chat.send_to_all("My chat message")

        :return class:`minetest.Chat`: chat object
        """
        return self._chat

    def log(self, line: str):
        """
        Write a line in the servers logfile.

        :param line: The log line
        :return: None
        """
        return self.lua.run('minetest.log("action", "{}")'.format(line))

    @property
    def players(self):
        """
        Get a list of all connected players on minetest server.

        :return: List of players
        """
        data = self.lua.run(
            """
            local players = {}
            for _,player in ipairs(minetest.get_connected_players()) do
                table.insert(players,player:get_player_name())
            end
            return players
            """
        )
        data = [] if len(data) == 0 else data
        return data

    def player(self, name: str):
        """
        Get a single players object.

        :param name: Name of the player
        :return: player object
        """
        return miney.Player(self, name)

    @property
    def node(self):
        """
        Manipulate and get information's about nodes.

        :return class:`minetest.Node`: Node manipulation functions
        """
        return self._node

    @property
    def lua(self):
        """
        Functions to run Lua inside Minetest.

        :return class:`minetest.Lua`: Lua related functions
        """
        return self._lua

    @property
    def time_of_day(self) -> int:
        """
        Get and set the time of the day between 0 and 1, where 0 stands for midnight, 0.5 for midday.

        :return: time of day as float.
        """
        return self.lua.run("return minetest.get_timeofday()")

    @time_of_day.setter
    def time_of_day(self, value: int):
        self.lua.run("return minetest.set_timeofday({})".format(value))

    @property
    def settings(self) -> dict:
        """
        Receive all server settings defined in "minetest.conf".

        :return: A dict with all non-default settings.
        """
        return self.lua.run("return minetest.settings:to_table()")

    def close(self) -> None:
        """
        Close the connection to the server.

        :return: None
        """
        self.connection.close()

    def __repr__(self):
        return '<minetest server "{}:{}">'.format(self.server, self.port)

    def __delete__(self, instance):
        self.connection.close()



import miney
from typing import Dict, List


class Node:
    """
    Manipulate and get information's about nodes.
    """
    def __init__(self, mt: miney.Minetest):
        self.mt = mt

    def __repr__(self):
        return '<minetest node functions>'

    def set(self, position: dict, node_name: str) -> None:
        """
        Set node at given position to "node_name" (something like "default:wood").
        You can get a list of all available nodes with :attr:`~pyminetest.Minetest.registered_nodes`

        :param position: a dict with x,y,z keys
        :param node_name: a valid node name
        :return: None
        """
        self.mt.lua.run("minetest.set_node({}, {{name=\"{}\"}})".format(self.mt.lua.dumps(position), node_name))

    def get(self, position: Dict):
        return self.mt.lua.run("return minetest.get_node({})".format(self.mt.lua.dumps(position)))

    @property
    def nodes(self) -> List:
        """
        Get a list of all available nodes/blocks.

        :return: List with all available node names
        """

        return self.mt.lua.run(
            """
            local nodes = {}
            for name, def in pairs(minetest.registered_nodes) do
                table.insert(nodes, name)
            end return nodes
            """
        )





"""
miney - The python interface to minetest
"""
from .minetest import Minetest
from .player import Player
from .chat import Chat
from .node import Node
from .lua import Lua
from .exceptions import *

name = "miney"



from typing import Dict
import miney


class Chat:
    """
    Chat functions.
    """
    def __init__(self, mt: miney.Minetest):
        self.mt = mt

    def __repr__(self):
        return '<minetest chat functions>'

    def send_to_all(self, message: str) -> None:
        """
        Send a chat message to all connected players.

        :param message: The chat message
        :return: None
        """
        self.mt.lua.run("minetest.chat_send_all('{}')".format(message.replace("\'", "\\'")))

    def send_to_player(self, playername: str, message: str):
        return self.mt.lua.run("return minetest.chat_send_player({}, {}})".format(playername, message))

    def format_message(self, playername: str, message: str):
        return self.mt.lua.run("return minetest.format_chat_message({}}, {}})".format(playername, message))

    def registered_commands(self):
        return self.mt.lua.run("return minetest.registered_chatcommands")

    def register_command(self, name, parameter: str, callback_function, description: str = "", privileges: Dict = None):

        if isinstance(callback_function, callable):
            pass
        elif isinstance(callback_function, str):
            pass

        return self.mt.lua.run(
            """
            return minetest.register_chatcommand(
                {name}, 
                {{params = {params}, description={description}, privs={privs}, func={func}}
            )""".format(
                name=name,
                params=parameter,
                description=description,
                privs=self.mt.lua.dumps(privileges) if privileges else "{}",
                func=callback_function
            )
        )

    def override_command(self, definition):
        pass

    def unregister_command(self, name: str):
        return self.mt.lua.run("return minetest.register_chatcommand({})".format(name))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="miney",
    version="0.1.1",
    author="Robert Lieback",
    author_email="robertlieback@zetabyte.de",
    description="The python interface to minetest",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/miney-py/miney",
    packages=["miney"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Topic :: Games/Entertainment"
    ],
    python_requires='>=3.6',
)



import re
import string
from random import choices
import miney


class Lua:
    """
    Lua specific functions.
    """
    def __init__(self, mt: miney.Minetest):
        self.mt = mt

    def run(self, lua_code: str):
        """
        Run load code on the minetest server.

        :param lua_code: lua code to run
        :return: The return value. Multiple values as a list.
        """
        # generates nearly unique id's (under 1000 collisions in 10 million values)
        result_id = ''.join(choices(string.ascii_lowercase + string.ascii_uppercase + string.digits, k=6))

        self.mt.send({"lua": lua_code, "id": result_id})

        try:
            return self.mt.receive(result_id)
        except miney.SessionReconnected:
            # We rerun the code, cause he was dropped during reconnect
            return self.run(lua_code)

    def dumps(self, data) -> str:
        """
        Convert python data types to a string with lua data types.

        :param data: Python data
        :return: Lua data
        """
        # credits:
        # https://stackoverflow.com/questions/54392760/serialize-a-dict-as-lua-table/54392761#54392761
        if type(data) is str:
            return '"{}"'.format(re.escape(data))
        if type(data) in (int, float):
            return '{}'.format(data)
        if type(data) is bool:
            return data and "true" or "false"
        if type(data) is list:
            l = "{"
            l += ", ".join([self.dumps(item) for item in data])
            l += "}"
            return l
        if type(data) is dict:
            t = "{"
            t += ", ".join(
                [
                    '[\"{}\"]={}'.format(re.escape(k), self.dumps(v)) for k, v in data.items()
                ]
            )
            t += "}"
            return t
        raise ValueError("Unknown type {}".format(type(data)))